"use strict";
/**
 * typescript-build-linker
 * Link together packages in a mono-repo
 */
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __spread = (this && this.__spread) || function () {
    for (var ar = [], i = 0; i < arguments.length; i++) ar = ar.concat(__read(arguments[i]));
    return ar;
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var fs = __importStar(require("fs"));
var path = __importStar(require("path"));
var tsconfig = __importStar(require("tsconfig"));
var globby_1 = __importDefault(require("globby"));
var zip_1 = __importDefault(require("@strong-roots-capital/zip"));
var fast_memoize_1 = __importDefault(require("fast-memoize"));
var functional_programming_1 = require("./functional-programming");
var utils_1 = require("./utils");
/* eslint-disable @typescript-eslint/no-var-requires */
var compose = require('just-compose');
/* eslint-enable @typescript-eslint/no-var-requires */
var traceDebug = functional_programming_1.traceDebugger(require('debug')('linker'));
// FIXME: is unsafe, wrap in an Either
// lernaPackages :: string -> Object
exports.readJson = fast_memoize_1.default(compose(functional_programming_1.id, // DISCUSS: why is this `id` necessary?
fs.readFileSync, JSON.parse.bind(null)));
// globbySync :: Glob[] -> File[]
var globbySync = function (globs) {
    return globby_1.default.sync(globs, { followSymlinkedDirectories: false });
};
// packageJsons :: Glob[] -> File[]
exports.packageJsons = fast_memoize_1.default(compose(globbySync, functional_programming_1.filter(function (filename) { return !filename.includes('node_modules'); }), functional_programming_1.filter(function (filename) { return filename.includes('package.json'); })));
// packageDirectories :: Glob[] -> File[]
exports.packageDirectories = compose(exports.packageJsons, functional_programming_1.map(path.dirname));
// packageNames :: Glob[] -> File[]
var packageNames = fast_memoize_1.default(compose(exports.packageJsons, functional_programming_1.map(exports.readJson), functional_programming_1.map(functional_programming_1.prop('name'))));
// jsonDependencies :: File -> string[]
var jsonDependencies = compose(exports.readJson, functional_programming_1.safeProp('dependencies'), Object.keys.bind(null));
// jsonDevDependencies :: File -> string[]
var jsonDevDependencies = compose(exports.readJson, functional_programming_1.safeProp('devDependencies'), Object.keys.bind(null));
var isInternalDependency = function (globs) { return function (dependency) {
    return packageNames(globs).includes(dependency);
}; };
// packageDependencies :: File[] -> string[][]
exports.packageDependencies = compose(exports.packageJsons, functional_programming_1.map(function (json) { return __spread(jsonDependencies(json), jsonDevDependencies(json)); }));
// internalPackageDependencies :: Glob[] -> File[][]
exports.internalPackageDependencies = function (globs) {
    return exports.packageDependencies(globs)
        .map(functional_programming_1.filter(isInternalDependency(globs)));
};
// :: Glob[] -> { [key: string]: File }
var internalPackages = function (glob) {
    return zip_1.default(packageNames(glob), exports.packageJsons(glob).map(path.dirname)).reduce(function (res, duple) {
        return (res[duple[0]] = duple[1], res);
    }, {});
};
// internalPackagePath :: Glob[] -> File -> File
exports.internalPackagePath = function (glob) { return function (pkg) {
    return internalPackages(glob)[pkg];
}; };
var relativePathFrom = function (from) { return function (to) {
    return path.relative(path.resolve(from), path.resolve(to));
}; };
// toPackageReferejces :: [File, File[]] -> [File, References]
exports.toPackageReferences = function (_a) {
    var _b = __read(_a, 2), pkg = _b[0], references = _b[1];
    return [
        pkg,
        {
            references: compose(functional_programming_1.map(relativePathFrom(pkg)), utils_1.sort, functional_programming_1.map(functional_programming_1.asProp('path')))(references)
        }
    ];
};
// FIXME: will not like when there is no existing tsconfig.json
var tsconfigParse = compose(functional_programming_1.id, // DISCUSS: is this necessary?
fs.readFileSync, utils_1.stringify, function (contents) { return tsconfig.parse(contents, ''); });
var writeJson = function (file, contents) {
    return fs.writeFileSync(file, JSON.stringify(contents, null, 4));
};
// FIXME: write more functionally and re-usably
// addReferencesToTsconfig :: [File, References -> [File, References
var addReferencesToTsconfig = function (_a) {
    var _b = __read(_a, 2), pkg = _b[0], references = _b[1];
    writeJson(path.join(pkg, 'tsconfig.json'), Object.assign(tsconfigParse(path.join(pkg, 'tsconfig.json')), references));
    return [pkg, references];
};
// writeReferences :: boolean -> [File, Reference[]] => [File, Reference[]]
exports.writePackageReferences = function (dryRun) {
    return compose(dryRun ? functional_programming_1.id : addReferencesToTsconfig);
};
// directoriesContainingPackages :: File[] -> File[]
exports.directoriesContainingPackages = compose(functional_programming_1.map(function (pkg) {
    return path.dirname(pkg)
        .split(path.sep)
        .scan(function (acc, dir) { return path.join(acc, dir); }, '')
        .filter(utils_1.isNotEmpty);
}), utils_1.flatten, utils_1.uniquify, traceDebug('directories containing packages'));
// childrenDirectories :: File -> File[] -> File[]]
var childrenDirectories = function (dir) {
    return compose(functional_programming_1.filter(utils_1.containedInDirectory(dir)), functional_programming_1.map(function (pkg) { return pkg.replace(dir + path.sep, ''); }), functional_programming_1.map(utils_1.split(path.sep)), functional_programming_1.map(utils_1.unsafeHead), utils_1.flatten, functional_programming_1.filter(utils_1.isNotEmpty), utils_1.uniquify);
};
// asParentReference :: File[] -> ParentReference
var asParentReference = function (packages) {
    return ({
        files: [],
        references: packages.map(functional_programming_1.asProp('path'))
    });
};
exports.toParentReferences = function (packages) { return function (parents) {
    return parents.mash(function (parent) { return [
        parent,
        compose(childrenDirectories(parent), asParentReference)(packages)
    ]; });
}; };
// FIXME: should _add_ references instead of overwrite entire tsconfig file
exports.writeParentReferences = function (dryRun) { return function (repositoryRoot) {
    return compose(dryRun
        ? functional_programming_1.id
        : functional_programming_1.objectMap(function (json, dir) {
            writeJson(path.join(repositoryRoot, dir, 'tsconfig.json'), json);
            return json;
        }));
}; };
